@@ -1,6 +1,6 @@
 
 /*
- * $Id: cache_cf.c,v 1.324.2.4 1999/04/19 00:14:34 wessels Exp $
+ * $Id: cache_cf.c,v 1.333.2.15 2000/06/13 14:19:57 wessels Exp $
  *
  * DEBUG: section 3     Configuration File Parsing
  * AUTHOR: Harvest Derived
@@ -12,10 +12,10 @@
  *  Internet community.  Development is led by Duane Wessels of the
  *  National Laboratory for Applied Network Research and funded by the
  *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
- *  Duane Wessels and the University of California San Diego.  Please
- *  see the COPYRIGHT file for full details.  Squid incorporates
- *  software developed and/or copyrighted by other sources.  Please see
- *  the CREDITS file for full details.
+ *  the Regents of the University of California.  Please see the
+ *  COPYRIGHT file for full details.  Squid incorporates software
+ *  developed and/or copyrighted by other sources.  Please see the
+ *  CREDITS file for full details.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -55,10 +55,8 @@ static const char *const B_MBYTES_STR = "MB";
 static const char *const B_GBYTES_STR = "GB";
 
 static const char *const list_sep = ", \t\n\r";
-
-static int http_header_first = 0;
-
-static void self_destruct(void);
+static int http_header_first;
+static int http_header_allowed = 0;
 
 static void configDoConfigure(void);
 static void parse_refreshpattern(refresh_t **);
@@ -78,8 +76,12 @@ static OBJH dump_config;
 static void dump_http_header(StoreEntry * entry, const char *name, HttpHeaderMask header);
 static void parse_http_header(HttpHeaderMask * header);
 static void free_http_header(HttpHeaderMask * header);
+static void parse_sockaddr_in_list(sockaddr_in_list **);
+static void dump_sockaddr_in_list(StoreEntry *, const char *, const sockaddr_in_list *);
+static void free_sockaddr_in_list(sockaddr_in_list **);
+static int check_null_sockaddr_in_list(const sockaddr_in_list *);
 
-static void
+void
 self_destruct(void)
 {
     fatalf("Bungled %s line %d: %s",
@@ -157,12 +159,17 @@ intlistFind(intlist * list, int i)
  * defined
  */
 
-#define GetInteger(var) \
-	token = strtok(NULL, w_space); \
-	if( token == NULL) \
-		self_destruct(); \
-	if (sscanf(token, "%d", &var) != 1) \
-		self_destruct();
+int
+GetInteger(void)
+{
+    char *token = strtok(NULL, w_space);
+    int i;
+    if (token == NULL)
+	self_destruct();
+    if (sscanf(token, "%d", &i) != 1)
+	self_destruct();
+    return i;
+}
 
 int
 parseConfigFile(const char *file_name)
@@ -181,6 +188,7 @@ parseConfigFile(const char *file_name)
 	cfg_filename = token + 1;
     memset(config_input_line, '\0', BUFSIZ);
     config_lineno = 0;
+    http_header_first = 0;
     while (fgets(config_input_line, BUFSIZ, fp)) {
 	config_lineno++;
 	if ((token = strchr(config_input_line, '\n')))
@@ -229,6 +237,7 @@ configDoConfigure(void)
 	Config.Announce.period = 86400 * 365;	/* one year */
 	Config.onoff.announce = 0;
     }
+#if USE_DNSSERVERS
     if (Config.dnsChildren < 1)
 	fatal("No dnsservers allocated");
     if (Config.dnsChildren > DefaultDnsChildrenMax) {
@@ -238,10 +247,11 @@ configDoConfigure(void)
 	    DefaultDnsChildrenMax);
 	Config.dnsChildren = DefaultDnsChildrenMax;
     }
+#endif
     if (Config.Program.redirect) {
 	if (Config.redirectChildren < 1) {
 	    Config.redirectChildren = 0;
-	    safe_free(Config.Program.redirect);
+	    wordlistDestroy(&Config.Program.redirect);
 	} else if (Config.redirectChildren > DefaultRedirectChildrenMax) {
 	    debug(3, 0) ("WARNING: redirect_children was set to a bad value: %d\n",
 		Config.redirectChildren);
@@ -271,13 +281,16 @@ configDoConfigure(void)
     if (Config.errHtmlText == NULL)
 	Config.errHtmlText = xstrdup(null_string);
     storeConfigure();
-    if (Config2.Accel.on && !strcmp(Config.Accel.host, "virtual"))
+    if (Config2.Accel.on && !strcmp(Config.Accel.host, "virtual")) {
 	vhost_mode = 1;
-    if (Config.Port.http == NULL)
+	if (Config.Accel.port == 0)
+	    vport_mode = 1;
+    }
+    if (Config.Sockaddr.http == NULL)
 	fatal("No http_port specified!");
     snprintf(ThisCache, sizeof(ThisCache), "%s:%d (%s)",
 	uniqueHostname(),
-	(int) Config.Port.http->i,
+	(int) ntohs(Config.Sockaddr.http->s.sin_port),
 	full_appname_string);
     /*
      * the extra space is for loop detection in client_side.c -- we search
@@ -285,7 +298,7 @@ configDoConfigure(void)
      */
     snprintf(ThisCache2, sizeof(ThisCache), " %s:%d (%s)",
 	uniqueHostname(),
-	(int) Config.Port.http->i,
+	(int) ntohs(Config.Sockaddr.http->s.sin_port),
 	full_appname_string);
     if (!Config.udpMaxHitObjsz || Config.udpMaxHitObjsz > SQUID_UDP_SO_SNDBUF)
 	Config.udpMaxHitObjsz = SQUID_UDP_SO_SNDBUF;
@@ -303,15 +316,21 @@ configDoConfigure(void)
 	debug(3, 0) ("WARNING: resetting 'maximum_single_addr_tries to 1\n");
 	Config.retry.maxtries = 1;
     }
+#if HEAP_REPLACEMENT
+    /* The non-LRU policies do not use referenceAge */
+#else
     if (Config.referenceAge < 300) {
 	debug(3, 0) ("WARNING: resetting 'reference_age' to 1 week\n");
 	Config.referenceAge = 86400 * 7;
     }
+#endif
     requirePathnameExists("MIME Config Table", Config.mimeTablePathname);
+#if USE_DNSSERVERS
     requirePathnameExists("cache_dns_program", Config.Program.dnsserver);
+#endif
     requirePathnameExists("unlinkd_program", Config.Program.unlinkd);
     if (Config.Program.redirect)
-	requirePathnameExists("redirect_program", Config.Program.redirect);
+	requirePathnameExists("redirect_program", Config.Program.redirect->key);
     if (Config.Program.authenticate)
 	requirePathnameExists("authenticate_program", Config.Program.authenticate->key);
     requirePathnameExists("Icon Directory", Config.icons.directory);
@@ -341,6 +360,33 @@ configDoConfigure(void)
 	Config.Wais.peer->host = xstrdup(Config.Wais.relayHost);
 	Config.Wais.peer->http_port = Config.Wais.relayPort;
     }
+    if (aclPurgeMethodInUse(Config.accessList.http))
+	Config2.onoff.enable_purge = 1;
+    if (NULL != Config.effectiveUser) {
+	struct passwd *pwd = getpwnam(Config.effectiveUser);
+	if (NULL == pwd)
+	    /*
+	     * Andres Kroonmaa <andre@online.ee>:
+	     * Some getpwnam() implementations (Solaris?) require
+	     * an available FD < 256 for opening a FILE* to the
+	     * passwd file.
+	     * DW:
+	     * This should be safe at startup, but might still fail
+	     * during reconfigure.
+	     */
+	    fatalf("getpwnam failed to find userid for effective user '%s'",
+		Config.effectiveUser,
+		xstrerror());
+	Config2.effectiveUserID = pwd->pw_uid;
+    }
+    if (NULL != Config.effectiveGroup) {
+	struct group *grp = getgrnam(Config.effectiveGroup);
+	if (NULL == grp)
+	    fatalf("getgrnam failed to find groupid for effective group '%s'",
+		Config.effectiveGroup,
+		xstrerror());
+	Config2.effectiveGroupID = grp->gr_gid;
+    }
 }
 
 /* Parse a time specification from the config file.  Store the
@@ -406,7 +452,7 @@ parseBytesLine(size_t * bptr, const char *units)
 	self_destruct();
     d = atof(token);
     m = u;			/* default to 'units' if none specified */
-    if (0 == d)
+    if (0.0 == d)
 	(void) 0;
     else if ((token = strtok(NULL, w_space)) == NULL)
 	debug(3, 0) ("WARNING: No units on '%s', assuming %f %s\n",
@@ -664,7 +710,7 @@ parse_delay_pool_rates(delayConfig * cfg)
 	if (sscanf(token, "%d", &i) != 1)
 	    self_destruct();
 	ptr->restore_bps = i;
-	GetInteger(i);
+	i = GetInteger();
 	ptr->max_bytes = i;
 	ptr++;
     }
@@ -698,7 +744,13 @@ parse_delay_pool_access(delayConfig * cfg)
 static void
 dump_http_header(StoreEntry * entry, const char *name, HttpHeaderMask header)
 {
-    storeAppendPrintf(entry, "%s\n", name);
+    int i;
+    for (i = 0; i < HDR_OTHER; i++) {
+	if (http_header_allowed && !CBIT_TEST(header, i))
+	    storeAppendPrintf(entry, "%s allow %s\n", name, httpHeaderNameById(i));
+	else if (!http_header_allowed && CBIT_TEST(header, i))
+	    storeAppendPrintf(entry, "%s deny %s\n", name, httpHeaderNameById(i));
+    }
 }
 
 static void
@@ -706,7 +758,6 @@ parse_http_header(HttpHeaderMask * header)
 {
     int allowed, id;
     char *t = NULL;
-
     if ((t = strtok(NULL, w_space)) == NULL) {
 	debug(3, 0) ("%s line %d: %s\n",
 	    cfg_filename, config_lineno, config_input_line);
@@ -723,16 +774,20 @@ parse_http_header(HttpHeaderMask * header)
 	debug(3, 0) ("parse_http_header: expecting 'allow' or 'deny', got '%s'.\n", t);
 	return;
     }
-
     if (!http_header_first) {
 	http_header_first = 1;
-	if (allowed)
+	if (allowed) {
+	    http_header_allowed = 1;
 	    httpHeaderMaskInit(header, 0xFF);
+	} else {
+	    http_header_allowed = 0;
+	    httpHeaderMaskInit(header, 0);
+	}
     }
     while ((t = strtok(NULL, w_space))) {
 	if ((id = httpHeaderIdByNameDef(t, strlen(t))) == -1)
-	    id = HDR_OTHER;
-	if (allowed)
+	    debug(3, 0) ("parse_http_header: Ignoring unknown header '%s'\n", t);
+	else if (allowed)
 	    CBIT_CLR(*header, id);
 	else
 	    CBIT_SET(*header, id);
@@ -752,12 +807,16 @@ dump_cachedir(StoreEntry * entry, const char *name, cacheSwap swap)
     int i;
     for (i = 0; i < swap.n_configured; i++) {
 	s = swap.swapDirs + i;
-	storeAppendPrintf(entry, "%s %s %d %d %d\n",
-	    name,
-	    s->path,
-	    s->max_size >> 10,
-	    s->l1,
-	    s->l2);
+	switch (s->type) {
+	case SWAPDIR_UFS:
+	case SWAPDIR_ASYNCUFS:
+	    storeUfsDirDump(entry, name, s);
+	    break;
+	default:
+	    debug(0, 0) ("dump_cachedir doesn't know about type %d\n",
+		(int) s->type);
+	    break;
+	}
     }
 }
 
@@ -773,71 +832,38 @@ check_null_string(char *s)
     return s == NULL;
 }
 
-static void
-parse_cachedir(cacheSwap * swap)
+void
+allocate_new_swapdir(cacheSwap * swap)
 {
-    char *token;
-    char *path;
-    int i;
-    int size;
-    int l1;
-    int l2;
-    unsigned int read_only = 0;
-    SwapDir *tmp = NULL;
-    if ((path = strtok(NULL, w_space)) == NULL)
-	self_destruct();
-    GetInteger(i);
-    size = i << 10;		/* Mbytes to kbytes */
-    if (size <= 0)
-	fatal("parse_cachedir: invalid size value");
-    GetInteger(i);
-    l1 = i;
-    if (l1 <= 0)
-	fatal("parse_cachedir: invalid level 1 directories value");
-    GetInteger(i);
-    l2 = i;
-    if (l2 <= 0)
-	fatal("parse_cachedir: invalid level 2 directories value");
-    if ((token = strtok(NULL, w_space)))
-	if (!strcasecmp(token, "read-only"))
-	    read_only = 1;
-    for (i = 0; i < swap->n_configured; i++) {
-	tmp = swap->swapDirs + i;
-	if (!strcmp(path, tmp->path)) {
-	    /* just reconfigure it */
-	    if (size == tmp->max_size)
-		debug(3, 1) ("Cache dir '%s' size remains unchanged at %d KB\n",
-		    path, size);
-	    else
-		debug(3, 1) ("Cache dir '%s' size changed to %d KB\n",
-		    path, size);
-	    tmp->max_size = size;
-	    if (tmp->flags.read_only != read_only)
-		debug(3, 1) ("Cache dir '%s' now %s\n",
-		    path, read_only ? "Read-Only" : "Read-Write");
-	    tmp->flags.read_only = read_only;
-	    return;
-	}
-    }
     if (swap->swapDirs == NULL) {
 	swap->n_allocated = 4;
 	swap->swapDirs = xcalloc(swap->n_allocated, sizeof(SwapDir));
     }
     if (swap->n_allocated == swap->n_configured) {
+	SwapDir *tmp;
 	swap->n_allocated <<= 1;
 	tmp = xcalloc(swap->n_allocated, sizeof(SwapDir));
 	xmemcpy(tmp, swap->swapDirs, swap->n_configured * sizeof(SwapDir));
 	xfree(swap->swapDirs);
 	swap->swapDirs = tmp;
     }
-    tmp = swap->swapDirs + swap->n_configured;
-    tmp->path = xstrdup(path);
-    tmp->max_size = size;
-    tmp->l1 = l1;
-    tmp->l2 = l2;
-    tmp->flags.read_only = read_only;
-    tmp->swaplog_fd = -1;
-    swap->n_configured++;
+}
+
+static void
+parse_cachedir(cacheSwap * swap)
+{
+    char *type_str;
+    if ((type_str = strtok(NULL, w_space)) == NULL)
+	self_destruct();
+    if (0 == strcasecmp(type_str, "ufs")) {
+	storeUfsDirParse(swap);
+#if USE_ASYNC_IO
+    } else if (0 == strcasecmp(type_str, "asyncufs")) {
+	storeAufsDirParse(swap);
+#endif
+    } else {
+	fatalf("Unknown cache_dir type '%s'\n", type_str);
+    }
 }
 
 static void
@@ -850,9 +876,15 @@ free_cachedir(cacheSwap * swap)
 	return;
     for (i = 0; i < swap->n_configured; i++) {
 	s = swap->swapDirs + i;
-	if (s->swaplog_fd > -1) {
-	    file_close(s->swaplog_fd);
-	    s->swaplog_fd = -1;
+	switch (s->type) {
+	case SWAPDIR_UFS:
+	case SWAPDIR_ASYNCUFS:
+	    storeUfsDirFree(s);
+	    break;
+	default:
+	    debug(0, 0) ("dump_cachedir doesn't know about type %d\n",
+		(int) s->type);
+	    break;
 	}
 	xfree(s->path);
 	filemapFreeMemory(s->map);
@@ -923,7 +955,7 @@ parse_peer(peer ** head)
     char *token = NULL;
     peer *p;
     int i;
-    ushortlist *u;
+    sockaddr_in_list *s;
     const char *me = null_string;	/* XXX */
     p = memAllocate(MEM_PEER);
     p->http_port = CACHE_HTTP_PORT;
@@ -936,13 +968,13 @@ parse_peer(peer ** head)
     if ((token = strtok(NULL, w_space)) == NULL)
 	self_destruct();
     p->type = parseNeighborType(token);
-    GetInteger(i);
+    i = GetInteger();
     p->http_port = (u_short) i;
-    GetInteger(i);
+    i = GetInteger();
     p->icp.port = (u_short) i;
     if (strcmp(p->host, me) == 0) {
-	for (u = Config.Port.http; u; u = u->next) {
-	    if (p->http_port != u->i)
+	for (s = Config.Sockaddr.http; s; s = s->next) {
+	    if (p->http_port != ntohs(s->s.sin_port))
 		continue;
 	    debug(15, 0) ("parse_peer: Peer looks like myself: %s %s/%d/%d\n",
 		p->type, p->host, p->http_port, p->icp.port);
@@ -991,6 +1023,12 @@ parse_peer(peer ** head)
 #endif
 	} else if (!strncasecmp(token, "login=", 6)) {
 	    p->login = xstrdup(token + 6);
+	} else if (!strncasecmp(token, "connect-timeout=", 16)) {
+	    p->connect_timeout = atoi(token + 16);
+#if USE_CACHE_DIGESTS
+	} else if (!strncasecmp(token, "digest-url=", 11)) {
+	    p->digest_url = xstrdup(token + 11);
+#endif
 	} else {
 	    debug(3, 0) ("parse_peer: token='%s'\n", token);
 	    self_destruct();
@@ -1001,11 +1039,14 @@ parse_peer(peer ** head)
     p->icp.version = ICP_VERSION_CURRENT;
     p->tcp_up = PEER_TCP_MAGIC_COUNT;
 #if USE_CARP
+#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> ((sizeof(u_long)*8)-(n))))
     if (p->carp.load_factor) {
 	/* calculate this peers hash for use in CARP */
 	p->carp.hash = 0;
 	for (token = p->host; *token != 0; token++)
-	    p->carp.hash += (p->carp.hash << 19) + *token;
+	    p->carp.hash += ROTATE_LEFT(p->carp.hash, 19) + *token;
+	p->carp.hash += p->carp.hash * 0x62531965;
+	p->carp.hash += ROTATE_LEFT(p->carp.hash, 21);
     }
 #endif
     /* This must preceed peerDigestCreate */
@@ -1190,6 +1231,7 @@ parse_hostdomaintype(void)
     }
 }
 
+#if UNUSED_CODE
 static void
 dump_ushortlist(StoreEntry * entry, const char *name, ushortlist * u)
 {
@@ -1233,6 +1275,7 @@ free_ushortlist(ushortlist ** P)
 	xfree(u);
     }
 }
+#endif
 
 static void
 dump_int(StoreEntry * entry, const char *name, int var)
@@ -1243,9 +1286,8 @@ dump_int(StoreEntry * entry, const char *name, int var)
 static void
 parse_int(int *var)
 {
-    char *token;
     int i;
-    GetInteger(i);
+    i = GetInteger();
     *var = i;
 }
 
@@ -1275,7 +1317,6 @@ parse_onoff(int *var)
 }
 
 #define free_onoff free_int
-#define free_httpanonymizer free_int
 #define dump_eol dump_string
 #define free_eol free_string
 
@@ -1336,11 +1377,11 @@ parse_refreshpattern(refresh_t ** head)
     if (token == NULL)
 	self_destruct();
     pattern = xstrdup(token);
-    GetInteger(i);		/* token: min */
+    i = GetInteger();		/* token: min */
     min = (time_t) (i * 60);	/* convert minutes to seconds */
-    GetInteger(i);		/* token: pct */
+    i = GetInteger();		/* token: pct */
     pct = (double) i / 100.0;
-    GetInteger(i);		/* token: max */
+    i = GetInteger();		/* token: max */
     max = (time_t) (i * 60);	/* convert minutes to seconds */
     /* Options */
     while ((token = strtok(NULL, w_space)) != NULL) {
@@ -1482,9 +1523,8 @@ dump_kb_size_t(StoreEntry * entry, const char *name, size_t var)
 static void
 parse_size_t(size_t * var)
 {
-    char *token;
     int i;
-    GetInteger(i);
+    i = GetInteger();
     *var = (size_t) i;
 }
 
@@ -1526,10 +1566,9 @@ free_ushort(u_short * u)
 static void
 parse_ushort(u_short * var)
 {
-    char *token;
     int i;
 
-    GetInteger(i);
+    i = GetInteger();
     if (i < 0)
 	i = 0;
     *var = (u_short) i;
@@ -1574,7 +1613,9 @@ parse_uri_whitespace(int *var)
     char *token = strtok(NULL, w_space);
     if (token == NULL)
 	self_destruct();
-    if (!strcasecmp(token, "deny"))
+    if (!strcasecmp(token, "strip"))
+	*var = URI_WHITESPACE_STRIP;
+    else if (!strcasecmp(token, "deny"))
 	*var = URI_WHITESPACE_DENY;
     else if (!strcasecmp(token, "allow"))
 	*var = URI_WHITESPACE_ALLOW;
@@ -1597,8 +1638,10 @@ dump_uri_whitespace(StoreEntry * entry, const char *name, int var)
 	s = "encode";
     else if (var == URI_WHITESPACE_CHOP)
 	s = "chop";
-    else
+    else if (var == URI_WHITESPACE_DENY)
 	s = "deny";
+    else
+	s = "strip";
     storeAppendPrintf(entry, "%s %s\n", name, s);
 }
 
@@ -1621,6 +1664,74 @@ parseNeighborType(const char *s)
     return PEER_SIBLING;
 }
 
+static void
+parse_sockaddr_in_list(sockaddr_in_list ** head)
+{
+    char *token;
+    char *t;
+    char *host;
+    const struct hostent *hp;
+    unsigned short port;
+    sockaddr_in_list *s;
+    while ((token = strtok(NULL, w_space))) {
+	host = NULL;
+	port = 0;
+	if ((t = strchr(token, ':'))) {
+	    /* host:port */
+	    host = token;
+	    *t = '\0';
+	    port = (unsigned short) atoi(t + 1);
+	    if (0 == port)
+		self_destruct();
+	} else if ((port = atoi(token)) > 0) {
+	    /* port */
+	} else {
+	    self_destruct();
+	}
+	s = xcalloc(1, sizeof(*s));
+	s->s.sin_port = htons(port);
+	if (NULL == host)
+	    s->s.sin_addr = any_addr;
+	else if (1 == safe_inet_addr(host, &s->s.sin_addr))
+	    (void) 0;
+	else if ((hp = gethostbyname(host)))	/* dont use ipcache */
+	    s->s.sin_addr = inaddrFromHostent(hp);
+	else
+	    self_destruct();
+	while (*head)
+	    head = &(*head)->next;
+	*head = s;
+    }
+}
+
+static void
+dump_sockaddr_in_list(StoreEntry * e, const char *n, const sockaddr_in_list * s)
+{
+    while (s) {
+	storeAppendPrintf(e, "%s %s:%d\n",
+	    n,
+	    inet_ntoa(s->s.sin_addr),
+	    ntohs(s->s.sin_port));
+	s = s->next;
+    }
+}
+
+static void
+free_sockaddr_in_list(sockaddr_in_list ** head)
+{
+    sockaddr_in_list *s;
+    while ((s = *head) != NULL) {
+	*head = s->next;
+	xfree(s);
+    }
+}
+
+static int
+check_null_sockaddr_in_list(const sockaddr_in_list * s)
+{
+    return NULL == s;
+}
+
 void
 configFreeMemory(void)
 {
