@@ -1,6 +1,6 @@
 
 /*
- * $Id: tools.c,v 1.173.2.4 1999/05/11 20:37:20 wessels Exp $
+ * $Id: tools.c,v 1.183.2.13 2000/03/24 20:55:31 wessels Exp $
  *
  * DEBUG: section 21    Misc Functions
  * AUTHOR: Harvest Derived
@@ -12,10 +12,10 @@
  *  Internet community.  Development is led by Duane Wessels of the
  *  National Laboratory for Applied Network Research and funded by the
  *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
- *  Duane Wessels and the University of California San Diego.  Please
- *  see the COPYRIGHT file for full details.  Squid incorporates
- *  software developed and/or copyrighted by other sources.  Please see
- *  the CREDITS file for full details.
+ *  the Regents of the University of California.  Please see the
+ *  COPYRIGHT file for full details.  Squid incorporates software
+ *  developed and/or copyrighted by other sources.  Please see the
+ *  CREDITS file for full details.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -48,14 +48,13 @@ Thanks!\n"
 static void fatal_common(const char *);
 static void fatalvf(const char *fmt, va_list args);
 static void mail_warranty(void);
-#if USE_ASYNC_IO
-static AIOCB safeunlinkComplete;
-#endif
 #if MEM_GEN_TRACE
 extern void log_trace_done();
 extern void log_trace_init(char *);
 #endif
 
+extern void (*failure_notify) (const char *);
+
 void
 releaseServerSockets(void)
 {
@@ -108,7 +107,7 @@ dumpMallocStats(void)
     fprintf(debug_log, "\tTotal free:            %6d KB %d%%\n",
 	(int) (ms.bytes_free >> 10),
 	percent(ms.bytes_free, ms.bytes_total));
-#elif HAVE_MALLINFO
+#elif HAVE_MALLINFO && HAVE_STRUCT_MALLINFO
     struct mallinfo mp;
     int t;
     if (!do_mallinfo)
@@ -191,7 +190,7 @@ rusage_maxrss(struct rusage *r)
     return r->ru_maxrss;
 #elif defined(BSD4_4)
     return r->ru_maxrss;
-#elif HAVE_GETPAGESIZE
+#elif defined(HAVE_GETPAGESIZE) && HAVE_GETPAGESIZE != 0
     return (r->ru_maxrss * getpagesize()) >> 10;
 #elif defined(PAGESIZE)
     return (r->ru_maxrss * PAGESIZE) >> 10;
@@ -216,7 +215,10 @@ PrintRusage(void)
 {
     struct rusage rusage;
     squid_getrusage(&rusage);
-    fprintf(debug_log, "CPU Usage: %.3f seconds\n", rusage_cputime(&rusage));
+    fprintf(debug_log, "CPU Usage: %.3f seconds = %.3f user + %.3f sys\n",
+	rusage_cputime(&rusage),
+	rusage.ru_utime.tv_sec + ((double) rusage.ru_utime.tv_usec / 1000000.0),
+	rusage.ru_stime.tv_sec + ((double) rusage.ru_stime.tv_usec / 1000000.0));
     fprintf(debug_log, "Maximum Resident Size: %d KB\n",
 	rusage_maxrss(&rusage));
     fprintf(debug_log, "Page faults with physical i/o: %d\n",
@@ -306,7 +308,7 @@ fatal_common(const char *message)
     syslog(LOG_ALERT, "%s", message);
 #endif
     fprintf(debug_log, "FATAL: %s\n", message);
-    if (opt_debug_stderr && debug_log != stderr)
+    if (opt_debug_stderr > 0 && debug_log != stderr)
 	fprintf(stderr, "FATAL: %s\n", message);
     fprintf(debug_log, "Squid Cache (Version %s): Terminated abnormally.\n",
 	version_string);
@@ -320,10 +322,10 @@ void
 fatal(const char *message)
 {
     releaseServerSockets();
-    /* check for store_rebuilding flag because fatal() is often
+    /* check for store_dirs_rebuilding because fatal() is often
      * used in early initialization phases, long before we ever
      * get to the store log. */
-    if (!store_rebuilding)
+    if (0 == store_dirs_rebuilding)
 	storeDirWriteCleanLogs(0);
     fatal_common(message);
     exit(shutting_down ? 0 : 1);
@@ -364,6 +366,7 @@ fatalvf(const char *fmt, va_list args)
 void
 fatal_dump(const char *message)
 {
+    failure_notify = NULL;
     releaseServerSockets();
     if (message)
 	fatal_common(message);
@@ -411,28 +414,49 @@ getMyHostname(void)
     LOCAL_ARRAY(char, host, SQUIDHOSTNAMELEN + 1);
     static int present = 0;
     const struct hostent *h = NULL;
-    char *t = NULL;
-
-    if ((t = Config.visibleHostname) != NULL)
-	return t;
-
-    /* Get the host name and store it in host to return */
-    if (!present) {
-	host[0] = '\0';
-	if (gethostname(host, SQUIDHOSTNAMELEN) == -1) {
-	    debug(50, 1) ("getMyHostname: gethostname failed: %s\n",
-		xstrerror());
-	    return NULL;
-	} else {
-	    if ((h = gethostbyname(host)) != NULL) {
-		/* DNS lookup successful */
-		/* use the official name from DNS lookup */
-		strcpy(host, h->h_name);
-	    }
+    if (Config.visibleHostname != NULL)
+	return Config.visibleHostname;
+    if (present)
+	return host;
+    host[0] = '\0';
+    if (Config.Sockaddr.http->s.sin_addr.s_addr != any_addr.s_addr) {
+	/*
+	 * If the first http_port address has a specific address, try a
+	 * reverse DNS lookup on it.
+	 */
+	h = gethostbyaddr((char *) &Config.Sockaddr.http->s.sin_addr,
+	    sizeof(Config.Sockaddr.http->s.sin_addr), AF_INET);
+	if (h != NULL) {
+	    /* DNS lookup successful */
+	    /* use the official name from DNS lookup */
+	    xstrncpy(host, h->h_name, SQUIDHOSTNAMELEN);
+	    debug(50, 4) ("getMyHostname: resolved %s to '%s'\n",
+		inet_ntoa(Config.Sockaddr.http->s.sin_addr),
+		host);
 	    present = 1;
+	    return host;
 	}
+	debug(50, 1) ("WARNING: failed to resolve %s to a hostname\n",
+	    inet_ntoa(Config.Sockaddr.http->s.sin_addr));
     }
-    return host;
+    /*
+     * Get the host name and store it in host to return
+     */
+    if (gethostname(host, SQUIDHOSTNAMELEN) < 0) {
+	debug(50, 1) ("WARNING: gethostname failed: %s\n", xstrerror());
+    } else if ((h = gethostbyname(host)) == NULL) {
+	debug(50, 1) ("WARNING: gethostbyname failed for %s\n", host);
+    } else {
+	debug(50, 6) ("getMyHostname: '%s' resolved into '%s'\n",
+	    host, h->h_name);
+	/* DNS lookup successful */
+	/* use the official name from DNS lookup */
+	xstrncpy(host, h->h_name, SQUIDHOSTNAMELEN);
+	present = 1;
+	return host;
+    }
+    fatal("Could not determine fully qualified hostname.  Please set 'visible_hostname'\n");
+    return NULL;		/* keep compiler happy */
 }
 
 const char *
@@ -444,31 +468,11 @@ uniqueHostname(void)
 void
 safeunlink(const char *s, int quiet)
 {
-#if USE_ASYNC_IO
-    aioUnlink(s,
-	quiet ? NULL : safeunlinkComplete,
-	quiet ? NULL : xstrdup(s));
-#else
     Counter.syscalls.disk.unlinks++;
     if (unlink(s) < 0 && !quiet)
 	debug(50, 1) ("safeunlink: Couldn't delete %s: %s\n", s, xstrerror());
-#endif
 }
 
-#if USE_ASYNC_IO
-static void
-safeunlinkComplete(int fd, void *data, int retcode, int errcode)
-{
-    char *s = data;
-    if (retcode < 0) {
-	errno = errcode;
-	debug(50, 1) ("safeunlink: Couldn't delete %s. %s\n", s, xstrerror());
-	errno = 0;
-    }
-    xfree(s);
-}
-#endif
-
 /* leave a privilegied section. (Give up any privilegies)
  * Routines that need privilegies can rap themselves in enter_suid()
  * and leave_suid()
@@ -477,34 +481,28 @@ safeunlinkComplete(int fd, void *data, int retcode, int errcode)
 void
 leave_suid(void)
 {
-    struct passwd *pwd = NULL;
-    struct group *grp = NULL;
     debug(21, 3) ("leave_suid: PID %d called\n", getpid());
     if (geteuid() != 0)
 	return;
     /* Started as a root, check suid option */
     if (Config.effectiveUser == NULL)
 	return;
-    if ((pwd = getpwnam(Config.effectiveUser)) == NULL)
-	return;
-    if (Config.effectiveGroup && (grp = getgrnam(Config.effectiveGroup))) {
-	if (setgid(grp->gr_gid) < 0)
-	    debug(50, 1) ("leave_suid: setgid: %s\n", xstrerror());
-    } else {
-	if (setgid(pwd->pw_gid) < 0)
-	    debug(50, 1) ("leave_suid: setgid: %s\n", xstrerror());
-    }
+#if HAVE_SETGROUPS
+    setgroups(1, &Config2.effectiveGroupID);
+#endif
+    if (setgid(Config2.effectiveGroupID) < 0)
+	debug(50, 0) ("ALERT: setgid: %s\n", xstrerror());
     debug(21, 3) ("leave_suid: PID %d giving up root, becoming '%s'\n",
-	getpid(), pwd->pw_name);
+	getpid(), Config.effectiveUser);
 #if HAVE_SETRESUID
-    if (setresuid(pwd->pw_uid, pwd->pw_uid, 0) < 0)
-	debug(50, 1) ("leave_suid: setresuid: %s\n", xstrerror());
+    if (setresuid(Config2.effectiveUserID, Config2.effectiveUserID, 0) < 0)
+	debug(50, 0) ("ALERT: setresuid: %s\n", xstrerror());
 #elif HAVE_SETEUID
-    if (seteuid(pwd->pw_uid) < 0)
-	debug(50, 1) ("leave_suid: seteuid: %s\n", xstrerror());
+    if (seteuid(Config2.effectiveUserID) < 0)
+	debug(50, 0) ("ALERT: seteuid: %s\n", xstrerror());
 #else
-    if (setuid(pwd->pw_uid) < 0)
-	debug(50, 1) ("leave_suid: setuid: %s\n", xstrerror());
+    if (setuid(Config2.effectiveUserID) < 0)
+	debug(50, 0) ("ALERT: setuid: %s\n", xstrerror());
 #endif
 }
 
@@ -553,7 +551,7 @@ writePidFile(void)
 	return;
     enter_suid();
     old_umask = umask(022);
-    fd = file_open(f, O_WRONLY | O_CREAT | O_TRUNC, NULL, NULL, NULL);
+    fd = file_open(f, O_WRONLY | O_CREAT | O_TRUNC);
     umask(old_umask);
     leave_suid();
     if (fd < 0) {
@@ -776,6 +774,7 @@ dlinkDelete(dlink_node * m, dlink_list * list)
 	list->head = m->next;
     if (m == list->tail)
 	list->tail = m->prev;
+    m->next = m->prev = NULL;
 }
 
 void
@@ -842,3 +841,58 @@ stringHasWhitespace(const char *s)
 {
     return (strcspn(s, w_space) != strlen(s));
 }
+
+void
+linklistPush(link_list ** L, void *p)
+{
+    link_list *l = xmalloc(sizeof(*l));
+    l->next = NULL;
+    l->ptr = p;
+    while (*L)
+	L = &(*L)->next;
+    *L = l;
+}
+
+void *
+linklistShift(link_list ** L)
+{
+    void *p;
+    link_list *l;
+    if (NULL == *L)
+	return NULL;
+    l = *L;
+    p = l->ptr;
+    *L = (*L)->next;
+    xfree(l);
+    return p;
+}
+
+
+/*
+ * Same as rename(2) but complains if something goes wrong;
+ * the caller is responsible for handing and explaining the 
+ * consequences of errors.
+ */
+int
+xrename(const char *from, const char *to)
+{
+    debug(21, 2) ("xrename: renaming %s to %s\n", from, to);
+    if (0 == rename(from, to))
+	return 0;
+    debug(21, errno == ENOENT ? 2 : 1) ("xrename: Cannot rename %s to %s: %s\n",
+	from, to, xstrerror());
+    return -1;
+}
+
+int
+stringHasCntl(const char *s)
+{
+    unsigned char c;
+    while ((c = (unsigned char) *s++) != '\0') {
+	if (c <= 0x1f)
+	    return 1;
+	if (c >= 0x7f && c <= 0x9f)
+	    return 1;
+    }
+    return 0;
+}
